from collections import Iterable
from collective.geo.contentlocations.interfaces import IGeoManager
from ftw.geo import _
from ftw.geo.interfaces import IGeocodableLocation
from geopy import geocoders
from plone.dexterity.interfaces import IDexterityContent
from plone.memoize import ram
from Products.Archetypes.interfaces import IBaseObject
from Products.statusmessages.interfaces import IStatusMessage
from urllib2 import URLError
from ZODB.POSException import ConflictError
from zope.annotation.interfaces import IAnnotations
from zope.component import queryAdapter
from zope.component.hooks import getSite
from zope.interface import alsoProvides
from zope.interface import Interface
from zope.interface import noLongerProvides


try:
    # geopy < 0.96
    from geopy.geocoders.googlev3 import GQueryError
    from geopy.geocoders.googlev3 import GTooManyQueriesError
    GeocoderQueryError = GQueryError
    GeocoderQuotaExceeded = GTooManyQueriesError
except ImportError:
    # geopy >= 0.96
    from geopy.exc import GeocoderQueryError
    from geopy.exc import GeocoderQuotaExceeded


LOCATION_KEY = 'ftw.geo.interfaces.IGeocodableLocation'


def display_status_message(msg):
    site = getSite()
    status = IStatusMessage(site.REQUEST)
    status.addStatusMessage(msg, type='info')


@ram.cache(lambda m, loc: loc)
def geocode_location(location):
    """Does a geocode lookup for `location` using the Google geocode API and
    returns a 3-tuple (place, coords, msg).

    If more than one result has been found, the first one is selected and
    `msg` will contain a status message telling the user which place has
    been selected.
    """
    msg = None

    # Google map api v3 does not take any api key
    # Check GoogleV3 implementation
    gmgeocoder = geocoders.GoogleV3()

    try:
        results = gmgeocoder.geocode(location, exactly_one=False)

        # geopy < 0.98 does not return a list in every case.
        if not isinstance(results, list):
            results = list(results)

        place, coords = results[0]
        if len(results) > 1:
            msg = _(u'msg_multiple_matches',
                    default=u'More than one location found, chose first match '
                    '"${place}". Please check that coordinates are correct.',
                    mapping=dict(place=place))
        return (place, coords, msg)

    except GeocoderQueryError:
        # Couldn't find a suitable location
        msg = _(u'msg_no_match',
                default=u'Couldn\'t find a suitable match for location '
                '"${location}". Please use the "coordinates" tab to manually '
                'set the correct map loaction.',
                mapping=dict(location=location.decode('utf-8')))
        display_status_message(msg)
        return

    except GeocoderQuotaExceeded:
        msg = _(u'msg_too_many_queries',
                default=u'Geocoding failed because daily query limit has '
                'been exceeded.')
        display_status_message(msg)
        return

    except URLError:
        msg = _(u'msg_network_error',
                default=u'Geocoding failed because of a network error.')
        display_status_message(msg)

    except ConflictError:
        raise

    except Exception, e:
        msg = _(u'msg_unhandled_exception',
                default=u'Geocoding failed because of an error: ${exception}',
                mapping=dict(exception=e))
        display_status_message(msg)
        return


class IGeoCoding(Interface):
    """Interface used on the request for preventing recursive firing the event.
    """


def at(obj, event):
    if IBaseObject.providedBy(obj):
        return geocodeAddressHandler(obj, event)
    else:
        return


def dx(obj, event):
    if IDexterityContent.providedBy(obj):
        return geocodeAddressHandler(obj, event)
    else:
        return


def geocodeAddressHandler(obj, event):
    """Handler to automatically do geocoding lookups for IGeoreferenceable
    objects that have an IGeocodableLocation adapter.
    """
    # When creating the obj, it has no request - we need to get it
    # from the site.
    request = getSite().REQUEST

    # When we run the gecode address handler it will fire an object modified
    # event - but we are subscribing to this event. This will result in a
    # recursion unless we prevent it.
    if IGeoCoding.providedBy(request):
        return
    alsoProvides(request, IGeoCoding)

    location_adapter = queryAdapter(obj, IGeocodableLocation)
    if not location_adapter:
        return

    location = location_adapter.getLocationString()
    geo_manager = queryAdapter(obj, IGeoManager)

    if location:
        ann = queryAdapter(obj, IAnnotations)
        previous_location = ann.get(LOCATION_KEY)
        # Only do the geocoding lookup if the location changed
        if not location == previous_location:
            geocoding_result = geocode_location(location)
            if geocoding_result:
                _place, coords, msg = geocoding_result
                if msg:
                    display_status_message(msg)
                geo_manager.setCoordinates('Point', (coords[1], coords[0]))
                # Update the stored location
                ann[LOCATION_KEY] = location

    elif not location and geo_manager:
        geo_manager.removeCoordinates()

    noLongerProvides(request, IGeoCoding)



from collective.geo.contentlocations.interfaces import IGeoManager
from collective.geo.geographer.interfaces import IGeoreferenceable
from ftw.geo.interfaces import IGeocodableLocation
from ftw.geo.testing import GEO_FUNCTIONAL_TESTING
from ftw.geo.tests.helpers import ExpectGeocodingRequest
from ftw.testbrowser import browsing
from ftw.testbrowser.pages import factoriesmenu
from ftw.testbrowser.pages import statusmessages
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import login
from plone.app.testing import setRoles
from plone.dexterity.content import Item
from plone.dexterity.fti import DexterityFTI
from plone.directives import form
from unittest2 import TestCase
from zope import schema
from zope.component import getSiteManager
from zope.interface import Interface
from zope.interface import implements
import transaction


class IAddress(Interface):
    pass


class IAddressSchema(form.Schema):

    address = schema.Text(title=u'Address', required=False)


class AddressLocationAdapter(object):

    def __init__(self, context):
        self.context = context

    def getLocationString(self):
        return self.context.address


class Address(Item):
    implements(IAddress, IGeoreferenceable)


class TestDexterityEvents(TestCase):

    layer = GEO_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        setRoles(self.portal, TEST_USER_ID, ['Manager'])
        login(self.portal, TEST_USER_NAME)

        # Because we have a bug if we use ftw.geo in combination
        # with the IReferenceable behavior, we run all tests with
        # this behavior.
        fti = DexterityFTI(
            u'Address',
            schema='.'.join((IAddressSchema.__module__, 'IAddressSchema')),
            klass='.'.join((Address.__module__, 'Address')),
            behaviors=['plone.app.referenceablebehavior.referenceable.IReferenceable'])

        self.portal.portal_types._setObject('Address', fti)
        fti.lookupSchema()

        getSiteManager().registerAdapter(factory=AddressLocationAdapter,
                                         required=[IAddress],
                                         provided=IGeocodableLocation)

        transaction.commit()

    def tearDown(self):
        getSiteManager().unregisterAdapter(factory=AddressLocationAdapter,
                                           required=[IAddress],
                                           provided=IGeocodableLocation)

    @browsing
    def test_geocoding_triggered_when_creating_object(self, browser):
        browser.login().open()
        factoriesmenu.add('Address')
        with ExpectGeocodingRequest():
            browser.fill({'Address': 'Bern, Switzerland'}).submit()
        statusmessages.assert_no_error_messages()

        obj = self.portal.get('address')
        self.assertEquals(('Point', (7.444608, 46.947922)),
                          IGeoManager(obj).getCoordinates())

    @browsing
    def test_geocoding_triggered_when_editing(self, browser):
        browser.login().open()
        factoriesmenu.add('Address')
        with ExpectGeocodingRequest():
            browser.fill({'Address': 'Bern, Switzerland'}).submit()
        statusmessages.assert_no_error_messages()
        obj = self.portal.get('address')

        browser.find('Edit').click()
        with ExpectGeocodingRequest('Zurich, Switzerland', (8.53918, 47.36864)):
            browser.fill({'Address': 'Zurich, Switzerland'}).submit()

        self.assertEquals(('Point', (47.36864, 8.53918)),
                          IGeoManager(obj).getCoordinates())

    @browsing
    def test_delete_geocoding_if_address_removed(self, browser):
        browser.login().open()
        factoriesmenu.add('Address')
        with ExpectGeocodingRequest():
            browser.fill({'Address': 'Bern, Switzerland'}).submit()
        statusmessages.assert_no_error_messages()

        obj = self.portal.get('address')
        self.assertEquals(('Point', (7.444608, 46.947922)),
                          IGeoManager(obj).getCoordinates())

        browser.find('Edit').click()
        browser.fill({'Address': ''}).submit()

        self.assertEquals((None, None), IGeoManager(obj).getCoordinates())



from Products.CMFCore.utils import getToolByName
import logging


log = logging.getLogger('ftw.geo')


def uninstall_products_maps(context):
    """Uninstalls Products.Maps using the quickinstaller tool.
    """
    quickinstaller = getToolByName(context, 'portal_quickinstaller')
    if quickinstaller.isProductInstalled('Maps'):
        log.info('Uninstalling Products.Maps...')
        quickinstaller.uninstallProducts(['Maps'])
        log.info('Successfully uninstalled Products.Maps.')


def import_various(context):
    """Import step for configuration that is not handled in xml files.
    """
    # Only run step if a flag file is present
    if context.readDataFile('ftw.geo_default.txt') is None:
        return
    site = context.getSite()
    uninstall_products_maps(site)



from ftw.testing.layer import ComponentRegistryLayer
from plone.app.testing import FunctionalTesting
from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import PloneSandboxLayer
from plone.app.testing import applyProfile
from plone.testing import zca
from zope.configuration import xmlconfig


class ZCMLLayer(ComponentRegistryLayer):
    """A layer which only sets up the zcml, but does not start a zope
    instance.
    """

    defaultBases = (zca.ZCML_DIRECTIVES,)

    def setUp(self):
        super(ZCMLLayer, self).setUp()
        import ftw.geo.tests
        self.load_zcml_file('test.zcml', ftw.geo.tests)


ZCML_LAYER = ZCMLLayer()


class GeoLayer(PloneSandboxLayer):

    defaultBases = (PLONE_FIXTURE, )

    def setUpZope(self, app, configurationContext):
        xmlconfig.string(
            '<configure xmlns="http://namespaces.zope.org/zope">'
            '  <include package="z3c.autoinclude" file="meta.zcml" />'
            '  <includePlugins package="plone" />'
            '  <includePluginsOverrides package="plone" />'
            '</configure>',
            context=configurationContext)


    def setUpPloneSite(self, portal):
        applyProfile(portal, 'plone.app.dexterity:default')


GEO_FIXTURE = GeoLayer()
GEO_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(GEO_FIXTURE, ),
    name="ftw.geo:functional")
